/*
 * txrxInt.h --
 *
 *    Declarations used internally by the Txrx extension.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) txrxInt.h 1.2 96/08/14 09:46:58 
 */

#ifndef _TXRXINT
#define _TXRXINT

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "txrx.h"

/*
 * Definitions of package and versions.
 */
#define TXRX_NAME "txrx"
#define TXRX_VERSION "1.0"
#define TXRX_MAJOR 1
#define TXRX_MINOR 0

/*
 * Predefined symbols, processed before any user defined data types.
 */
#define RPCV2DEF "\
union __auth switch (enum {AUTH_NONE = 0, AUTH_SYS = 1, AUTH_SHORT = 2, \
  AUTH_DH = 3, AUTH_KERB4 = 4} auth_flavor) {\
case AUTH_NONE:\
 opaque body<400>;\
case AUTH_SYS:\
 struct { \
  int length_of_auth_sys;\
  int stamp;\
  string machinename<255>;\
  int uid;\
  int gid;\
  int gids<16>;\
 } authsys_parms;\
default:\
  opaque default_body<400>;\
};\
\
struct __rpc {\
 unsigned int xid;\
 union switch (enum {CALL = 0, REPLY = 1} msg_type) {\
  case CALL:\
   struct {\
    unsigned int rpcvers; \
    unsigned int prog;\
    unsigned int vers;\
    unsigned int proc;\
    __auth cred;\
    __auth verf;\
   } call_body;\
  case REPLY:\
   union switch (enum {MSG_ACCEPTED = 0, MSG_DENIED = 1} reply_stat) {\
    case MSG_ACCEPTED:\
     struct {\
      __auth verf_reply;\
      union switch (enum {SUCCESS = 0, PROG_UNAVAIL = 1,\
       PROG_MISMATCH = 2, PROC_UNAVAIL = 3,\
       GARBAGE_ARGS = 4, SYSTEM_ERR = 5} acc) {\
       case SUCCESS:\
        opaque results[0];\
       case PROG_MISMATCH:\
        struct {\
         unsigned int low;\
         unsigned int high;\
        } mismatch_info;\
       default:\
        /* PROG_UNAVAIL, PROC_UNAVAIL, GARBAGE_ARGS, SYSTEM_ERR. */\
        void;\
      } reply_data;\
     } accepted_reply;\
    case MSG_DENIED:\
     union switch (enum {RPC_MISMATCH = 0, AUTH_ERROR = 1} reject) {\
      case RPC_MISMATCH:\
       struct {\
        unsigned int low;\
        unsigned int high;\
       } mismatch_info;\
      case AUTH_ERROR:\
       enum {AUTH_OK = 0, AUTH_BADCRED = 1, AUTH_REJECTEDCRED = 2,\
        AUTH_BADVERF = 3, AUTH_REJECTEDVERF = 4, \
        AUTH_TOOWEAK = 5, AUTH_INVALIDRESP  = 6, \
        AUTH_FAILED = 7, AUTH_KERB_GENERIC = 8,\
        AUTH_TIMEEXPIRE = 9, AUTH_TKT_FILE = 10, \
        AUTH_DECODE = 11, AUTH_NET_ADDR = 12} auth_stat;\
    } rejected_reply;\
   } reply_body;\
  } call_reply;\
};"
/*
 * Defintions of symbols output by the lexical analizer and consumed
 * by the parser.
 */
#define UNKNOWN               0
#define LPAREN                1
#define RPAREN                2
#define LSQUARE               3
#define RSQUARE               4
#define LANGLE                5
#define RANGLE                6
#define LCURLY                7
#define RCURLY                8
#define COLUMN                9
#define SEMICOLUMN            10
#define STAR                  11
#define EQUAL                 12
#define MINUS                 13
#define COMMA                 14

#define VALUE                 15
#define IDENT                 16
#define CONSTVAL              17

#define KEYWRD                18
#define EOFFOUND              19

/*
 * Constant numeric values have maximum length MAXVALUELEN in characters.
 * This constant is implementation dependent - it depends on the maximum size
 * of an integer that can be stored in a double without precision loss.
 */
#define MAXVALUELEN           16

/*
 * Constants used to delimit a tape symbol.
 */
#define CELL_START            0xbe1ea000
#define CELL_END              0x000ea1be
/*
 * Entry point for the first executable instruction of the interpreter.
 * Increment for code memory allocation and initial code allocation size,
 * must be larger than CODE_START.
 */
#define CODE_START            256
#define CODE_INCREMENT        270

/*
 * Network ordering of bytes is always big endian.
 * Computer ordering of bytes may be little endian.
 */
#define TXRX_BIGENDIAN        0
#define TXRX_LITTLEENDIAN     1

/*
 * Increment for data structure allocation.
 */
#define DATA_INCREMENT        1024

/*
 * Definition of the data structure storing the state of the package.
 */
typedef struct {
    Tcl_HashTable tapesTable;   /* Hash table with bindings between 
                                 * tapes and tape (command) names.
                                 */
    int tapeId;                 /* Counter used to generate names.         */
    Tcl_Interp *interp;         /* Interpreter for this extenstion.        */
    Tcl_Command command;        /* TXRX_NAME command pointer.              */
    int isSafe;                 /* Flag set indicates interpreter is safe. */
} TxrxData;

/*
 * Data structures used to store a constant or a type description.
 */
typedef enum {
    SYM_CONST = 0, SYM_ENUM = 1, SYM_STRUCT = 2, SYM_UNION = 3, 
    SYM_PROGRAM = 4, SYM_VERSION = 5, SYM_PROCEDURE = 6, 
    SYM_STRING = 7, SYM_OPAQUE = 8, 
    SYM_INT = 9, SYM_UNSIGNED = 10, SYM_BOOL = 11, 
    SYM_HYPER = 12, SYM_UNSIGNEDHYPER = 13, 
    SYM_FLOAT = 14, SYM_DOUBLE = 15, SYM_QUADRUPLE = 16,
    SYM_VOID = 17, SYM_ARRAY = 18, SYM_POINTER = 19, SYM_TYPE = 20
} SymbolType;

/*
 * Note: to make type analysis simpler and translation 
 * more efficient it would be great if predefined types
 * like int, double, string<>, netobj, bool, long, etc
 * would have symbols in the symbol table from the beginning.
 * This creates the possibility for the user too send
 * elementary data types without having to (re)define
 * the types in the protocol file.
 */
/*
 * Symbol table entry. The type field decides the contents of the
 * other fields. The symbol name is not stored as symbol data as all
 * hash tables in the program have the name of the symbol as key:

 *  - CONST symbols the constValue is filled with the actual value
 * of the constant (fits in a long int). When a const symbol 
 * has the name of a procedure it keeps a pointer to the procedure
 * in referenceValue.

 *  - ENUM symbols have a hashTable with the equivalences between
 * names of identifiers and having as values symbols of type CONST.
 * The number of distinct elements in an enum is stored in constValue.

 *  - STRUCT symbols have a hashTable with the keys "0" to "n"
 *  representing the fields of a struct. The symbols associated
 *  are describing the type of each field. Similar to the PROCEDURE
 *  symbol the number of fields is stored in constValue.

 *  - UNION symbols store in the hashTable as keys the values (numeric)
 * and the CaseData structures - corresponding to the value.
 * If a default case declaration exists it is stored in defaultValue
 * The discriminant declaration is stored in referenceValue.

 *  - STRING/OPAQUE symbols store in constValue the size of the data
 * structure and a non-NULL identValue means variable length.

 *  - INT et al., including void, symbols do not store any information 
 * i.e. all pointers are NULL.

 *  - ARRAY symbols store in constValue the size of the array and in 
 * referenceValue the actual component symbol. The hashTable field is 
 * not used. identValue not NULL means variable length array.

 *  - POINTER symbols store in referenceValue the symbol for the type
 * pointed at.

 *  - TYPE stores in identValue the name of the type or a pointer to
 * another type.
 *
 *  - PROGRAM and VERSION store in hash table version respectively
 * procedure symbols. constValue stores the program number respectively
 * the version number. Version symbols store a pointer to the corresponding
 * program in referenceValue.
 *
 *  - PROCEDURE stores in the hash table with string keys of the format
 *  "n" where n starts at 0 the type symbols of the arguments. Key "0"
 *  is used to store the type of the result of the call. As for the
 *  version definition, constValue stores the number of the procedure.
 *  The number of arguments is kept casted to a char * in identValue.
 *  A pointer to the version is stored in referenceValue.
 */
/* See above for the values stored in a symbol. */
typedef struct symbolData {
    SymbolType type;
    long constValue;
    char *identValue;
    struct symbolData *referenceValue;
    char *defaultValue;
    Tcl_HashTable *hashTablePtr;
    int codeAddress;              /* Location on tape where
                                   * the code for the symbol begins.
                                   */
} SymbolData;

/*
 * Union case description.
 */
typedef struct caseData {
    SymbolData *symbolPtr;
    char symbolNamePtr[MAXIDENTLEN + 1];
    Token label;
} CaseData;

/*
 * Tape data structures. A tape consists of a sequence of variable
 * length cells, each cell describing an XDR/RPC data type.
 */
typedef enum {  
    TIMPOSSIBLE = 0,
    TVOID = 1, TINT = 2, TUINT = 3, THYPER = 4, TUHYPER = 5,
    TFLOAT = 6, TDOUBLE = 7, TQUAD = 8, TBOOL = 9,
    TENUM = 64, TOPAQUE = 65, TSTRING = 66,
    TARRAY = 67, TSTRUCT = 68, TUNION = 69,
    TPOINTER = 70,
    TPROGRAM = 71, TVERSION = 72, TPROC = 73,
    TMODULE = 74
} CellType;

typedef struct {
    long numValues;
    long values[1];
} EnumType;

typedef struct {
    long length;
    long isVariable;
} OpaqueType;

typedef struct {
    long length;
    long isVariable;
} StringType;

typedef struct {
    long length;
    long isVariable;
    long location;
} ArrayType;

typedef struct {
    long numElements;
    long locations[1];
} StructType;

typedef struct {
    long discriminantLocation;
    long numElements;
    long hasDefault;
    long defaultLocation;
    struct {
        long value;
        long location;
    } choices[1];
} UnionType;

typedef struct {
    long location;
} PointerType;

typedef struct {
    long numElements;
    long numProgram;
    struct {
        long version;
        long location;
    } choices[1];
} ProgramType;

typedef struct {
    long numElements;
    long numVersion;
    struct {
        long number;
        long location;
    } choices[1];
} VersionType;

typedef struct {
    long numParameters;
    long numProcedure;
    long locations[1];
} ProcType;

typedef struct {
    long numPrograms;
    struct {
        long number;
        long location;
    } choices[1];
} ModuleType;

typedef struct {
    long                cellStart;       /* Predefined constant starting a
                                          * symbol.
                                          */
    CellType            type;            /* Type of the cell.              */
    long                clientData;      /* Usually it's a pointer to the
                                          * symbol.
                                          */
    union {
        EnumType        enumType;
        OpaqueType      opaqueType;
        StringType      stringType;
        ArrayType       arrayType;
        StructType      structType;
        UnionType       unionType;
        PointerType     pointerType;
        ProgramType     programType;
        VersionType     versionType;
        ProcType        procType;
        ModuleType      moduleType;
    } data;
    long                cellEnd;          /* Constant marking an end-cell. */
} Cell;

/*
 * A tape consists of an array of cells, each cell formated as it's
 * type requires. The first 512 bytes of a tape are reserved for
 * internal purposes. The first 256 bytes of a tape can be used to store
 * identification information because any offset value < 64 is interpreted
 * as a basic type, not as a location on the tape. This provides
 * inlining for basic types like int, double, bool, etc. Frequently
 * used types like string<> or opaque<> might also be considered
 * for inclusion among the basic types.
 *
 * Note: 1 offset unit (the contents of location fields) has a length of
 * 4 bytes. No byte addressing is possible. This wastes space but most
 * protocols are brief.
 * 
 */

/*
 * Procedures used internally by TxRx and shared between files.
 */

EXTERN int                 Txrx_Compile _ANSI_ARGS_((Txrx_Tape *tapePtr));

EXTERN int                 Txrx_PrintSymbols _ANSI_ARGS_((Txrx_Tape *tapePtr));
EXTERN int                 Txrx_PrintTape _ANSI_ARGS_((Tcl_Interp *interp,
                                                       Txrx_Tape *tapePtr));
EXTERN int                 Txrx_FindProc _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                      char *namePtr,
                                                      int *addrPtr));
EXTERN int                 Txrx_InfoProc _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                      int addr));

EXTERN SymbolData *        Txrx_GetSymbol _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                       char *typeName));

EXTERN int                 Txrx_ToXDR _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                    char *listPtr,
                                                    int addr));
EXTERN int                 Txrx_ToList _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                     int addr));
#endif /* _TXRXINT */
